import lightning
import pytest

from pvnet_summation.data.datamodule import SavedPredictionDataModule


@pytest.fixture()
def saved_prediction_datamodule(presaved_predictions_dir):
    return SavedPredictionDataModule(
        sample_dir=presaved_predictions_dir,
        batch_size=2,
        num_workers=0,
        prefetch_factor=None,
    )


def test_model_trainer_fit(model, saved_prediction_datamodule):
    trainer = lightning.pytorch.trainer.trainer.Trainer(fast_dev_run=True)
    trainer.fit(model=model, datamodule=saved_prediction_datamodule)



"""Simple model which only uses outputs of PVNet for all GSPs"""

from typing import Optional

import numpy as np
import pvnet
import torch
import torch.nn.functional as F
from pvnet.models.multimodal.linear_networks.basic_blocks import AbstractLinearNetwork
from pvnet.optimizers import AbstractOptimizer
from torch import nn

from pvnet_summation.models.base_model import BaseModel


class FlatModel(BaseModel):
    """Neural network which combines GSP predictions from PVNet naively

    This model flattens all the features into a 1D vector before feeding them into the sub network
    """

    name = "FlatModel"

    def __init__(
        self,
        output_network: AbstractLinearNetwork,
        model_name: str,
        model_version: Optional[str] = None,
        num_locations: int = 317,
        output_quantiles: Optional[list[float]] = None,
        relative_scale_pvnet_outputs: bool = False,
        predict_difference_from_sum: bool = False,
        optimizer: AbstractOptimizer = pvnet.optimizers.Adam(),
    ):
        """Neural network which combines GSP predictions from PVNet naively

        Args:
            output_network: A partially instantiated pytorch Module class used to combine the 1D
                features to produce the forecast.
            model_name: Model path either locally or on huggingface.
            model_version: Model version if using huggingface. Set to None if using local.
            num_locations: The number of regional GSP locations.
            output_quantiles: A list of float (0.0, 1.0) quantiles to predict values for. If set to
                None the output is a single value.
            relative_scale_pvnet_outputs: If true, the PVNet predictions are scaled by a factor
                which is proportional to their capacities.
            predict_difference_from_sum: Whether to use the sum of GSPs as an estimate for the
                national sum and train the model to correct this estimate. Otherwise the model tries
                to learn the national sum from the PVNet outputs directly.
            optimizer (AbstractOptimizer): Optimizer
        """

        super().__init__(model_name, model_version, num_locations, optimizer, output_quantiles)

        self.relative_scale_pvnet_outputs = relative_scale_pvnet_outputs
        self.predict_difference_from_sum = predict_difference_from_sum

        self.model = output_network(
            in_features=np.prod(self.pvnet_output_shape),
            out_features=self.num_output_features,
        )

        # Add linear layer if predicting difference from sum
        # This allows difference to be positive or negative
        if predict_difference_from_sum:
            self.model = nn.Sequential(
                self.model, nn.Linear(self.num_output_features, self.num_output_features)
            )

        self.save_hyperparameters()

    def forward(self, x):
        """Run model forward"""

        if "pvnet_outputs" not in x:
            x["pvnet_outputs"] = self.predict_pvnet_batch(x["pvnet_inputs"])

        if self.relative_scale_pvnet_outputs:
            eff_cap = x["effective_capacity"].float().unsqueeze(-1)

            if self.pvnet_model.use_quantile_regression:
                eff_cap = eff_cap.unsqueeze(-1)

            # The effective_capacit[ies] are relative fractions of the national capacity. They sum
            # to 1 and they are quite small values. For the largest GSP the capacity is around 0.03.
            # Therefore we apply this scaling to make the input values a more sensible size
            x_in = x["pvnet_outputs"] * eff_cap * 100
        else:
            x_in = x["pvnet_outputs"]

        x_in = torch.flatten(x_in, start_dim=1)
        out = self.model(x_in)

        if self.use_quantile_regression:
            # Shape: batch_size, seq_length * num_quantiles
            out = out.reshape(out.shape[0], self.forecast_len, len(self.output_quantiles))

        if self.predict_difference_from_sum:
            gsp_sum = self.sum_of_gsps(x)

            if self.use_quantile_regression:
                gsp_sum = gsp_sum.unsqueeze(-1)

            out = F.leaky_relu(gsp_sum + out)

        return out



"""Pytorch lightning datamodules for loading pre-saved samples and predictions.

The pre-saced samplea can be prepared using the save_concurrent_samples.py script from the PVNet
library: https://github.com/openclimatefix/PVNet
"""

from glob import glob
import torch
from torch.utils.data import Dataset, DataLoader, default_collate
from lightning.pytorch import LightningDataModule
from ocf_data_sampler.load.gsp import open_gsp


# https://github.com/pytorch/pytorch/issues/973
torch.multiprocessing.set_sharing_strategy("file_system")


def collate_summation_samples(samples: list[dict]) -> dict:
    batch_dict = {}
    batch_dict["pvnet_inputs"] = [s["pvnet_inputs"] for s in samples]
    for key in ["effective_capacity", "national_targets", "times"]:
        batch_dict[key] = torch.stack([s[key] for s in samples])
    return batch_dict


def get_national_outturns(gsp_data, times) -> torch.Tensor:
    return torch.as_tensor(
        gsp_data.sel(time_utc=times.cpu().numpy().astype("datetime64[ns]")).values
    )


def get_sample_valid_times(sample: dict) -> torch.Tensor:
    id0 = int(sample["gsp_t0_idx"])
    return sample["gsp_time_utc"][0, id0 + 1 :]


def get_sample_capacities(sample: dict) -> torch.Tensor:
    return sample["gsp_effective_capacity_mwp"].float().unsqueeze(-1)


class SavedSampleDataset(Dataset):
    def __init__(self, sample_dir: str, gsp_zarr_path: str, gsp_boundaries_version="20220314"):
        self.sample_filepaths = glob(f"{sample_dir}/*.pt")

        # Load and nornmalise the national GSP data to use as target values
        gsp_data = (
            open_gsp(zarr_path=gsp_zarr_path, boundaries_version=gsp_boundaries_version)
            .sel(gsp_id=0)
            .compute()
        )
        gsp_data = gsp_data / gsp_data.effective_capacity_mwp

        self.gsp_data = gsp_data

    def __len__(self) -> int:
        return len(self.sample_filepaths)

    def __getitem__(self, idx) -> dict:
        sample = torch.load(self.sample_filepaths[idx], weights_only=False)

        sample_valid_times = get_sample_valid_times(sample)

        national_outturns = get_national_outturns(self.gsp_data, sample_valid_times)

        national_capacity = get_national_outturns(
            self.gsp_data.effective_capacity_mwp, sample_valid_times
        )[0]

        gsp_capacities = get_sample_capacities(sample)

        gsp_relative_capacities = gsp_capacities / national_capacity

        return dict(
            pvnet_inputs=sample,
            effective_capacity=gsp_relative_capacities,
            national_targets=national_outturns,
            times=sample_valid_times,
        )


class SavedSampleDataModule(LightningDataModule):
    """Datamodule for training pvnet_summation."""

    def __init__(
        self,
        sample_dir: str,
        gsp_zarr_path: str,
        gsp_boundaries_version: str = "20220314",
        batch_size: int = 16,
        num_workers: int = 0,
        prefetch_factor: int | None = None,
    ):
        """Datamodule for training pvnet_summation.

        Args:
            sample_dir: Path to the directory of pre-saved samples.
            gsp_zarr_path: Path to zarr file containing GSP ID 0 outputs
            batch_size: Batch size.
            num_workers: Number of workers to use in multiprocess batch loading.
            prefetch_factor: Number of data will be prefetched at the end of each worker process.
        """
        super().__init__()
        self.gsp_zarr_path = gsp_zarr_path
        self.gsp_boundaries_version = gsp_boundaries_version
        self.sample_dir = sample_dir
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.prefetch_factor = prefetch_factor

    @property
    def _dataloader_kwargs(self) -> dict:
        return dict(
            batch_size=self.batch_size,
            sampler=None,
            batch_sampler=None,
            num_workers=self.num_workers,
            collate_fn=None if self.batch_size is None else collate_summation_samples,
            pin_memory=False,
            drop_last=False,
            timeout=0,
            worker_init_fn=None,
            prefetch_factor=self.prefetch_factor,
            persistent_workers=self.num_workers > 0,
        )

    def train_dataloader(self, shuffle: bool = False) -> DataLoader:
        """Construct train dataloader"""
        dataset = SavedSampleDataset(
            f"{self.sample_dir}/train",
            self.gsp_zarr_path,
            self.gsp_boundaries_version,
        )
        return DataLoader(dataset, shuffle=shuffle, **self._dataloader_kwargs)

    def val_dataloader(self, shuffle: bool = False) -> DataLoader:
        """Construct val dataloader"""
        dataset = SavedSampleDataset(
            f"{self.sample_dir}/val",
            self.gsp_zarr_path,
            self.gsp_boundaries_version,
        )
        return DataLoader(dataset, shuffle=shuffle, **self._dataloader_kwargs)


class SavedPredictionDataset(Dataset):
    def __init__(self, sample_dir: str):
        self.sample_filepaths = sorted(glob(f"{sample_dir}/*.pt"))

    def __len__(self) -> int:
        return len(self.sample_filepaths)

    def __getitem__(self, idx: int) -> dict:
        return torch.load(self.sample_filepaths[idx], weights_only=False)


class SavedPredictionDataModule(LightningDataModule):
    """Datamodule for loading pre-saved PVNet predictions to train pvnet_summation."""

    def __init__(
        self,
        sample_dir: str,
        batch_size: int = 16,
        num_workers: int = 0,
        prefetch_factor: int | None = None,
    ):
        """Datamodule for loading pre-saved PVNet predictions to train pvnet_summation.

        Args:
            sample_dir: Path to the directory of pre-saved batches.
            batch_size: Batch size.
            num_workers: Number of workers to use in multiprocess batch loading.
            prefetch_factor: Number of data will be prefetched at the end of each worker process.
        """
        super().__init__()
        self.sample_dir = sample_dir
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.prefetch_factor = prefetch_factor

    @property
    def _dataloader_kwargs(self) -> dict:
        return dict(
            batch_size=self.batch_size,
            sampler=None,
            batch_sampler=None,
            num_workers=self.num_workers,
            collate_fn=None if self.batch_size is None else default_collate,
            pin_memory=False,
            drop_last=False,
            timeout=0,
            worker_init_fn=None,
            prefetch_factor=self.prefetch_factor,
            persistent_workers=self.num_workers > 0,
        )

    def train_dataloader(self, shuffle: bool = True) -> DataLoader:
        """Construct train dataloader"""
        dataset = SavedPredictionDataset(f"{self.sample_dir}/train")
        return DataLoader(dataset, shuffle=shuffle, **self._dataloader_kwargs)

    def val_dataloader(self, shuffle: bool = False) -> DataLoader:
        """Construct val dataloader"""
        dataset = SavedPredictionDataset(f"{self.sample_dir}/val")
        return DataLoader(dataset, shuffle=shuffle, **self._dataloader_kwargs)
